--[[

	VLC Addon: TvTeam,
		version 1.0.0,
			2017-06-22

--]]


-- importing required modues

date = require("date")
json = require("dkjson")
md5 = require("md5")

-- global variables

API_URL = "https://tv.team/api/"
FILE_CONFIG = "tvteam_config"
FILE_SESSION = "tvteam_session"

IS_CONFIG_DIALOG_OPEN = false
IS_SESSION_OPENED = false

TOKENS = {}

ITEM_TO_REMOVE = 0

TIMEZONE_IN_API_RESPONSE = "+03:00"

dlg = nil


-- description of the addon

function descriptor()

   return {
      title = "TvTeam IPTV",
      version = "1.0",
      author = "TvTeam",
      url = 'http://tv.team',
      shortdesc = "TvTeam IPTV",
      description = "Addon for watching TvTeam live and archive channels",
      capabilities = { }
   }
   
end

-- gets executed after the initial launch

function activate()	
	
	-- will display config dialog first if new session couldn't be started
	if not check_account() then
		create_dialog_config()
	else -- seems like everything is fine, so we can start by opening the main window
		create_dialog_main()
	end
	
end


--- gets executed upon the deactivation

function deactivate()

end


-- callback for closing the dialog

function close()

	-- if session is opened and user is in config window, then by closing it, we return them back to the main one
	if (IS_CONFIG_DIALOG_OPEN == true and IS_SESSION_OPENED == true) then
		close_dialog()
		create_dialog_main()
	else
		vlc.deactivate()
	end
   
end


-- helper function to close current dialog in a clean way

function close_dialog()

	if dlg ~= nil then 
		dlg:hide() 
	end

	dlg = nil
	collectgarbage()
  
end


-- getting rid of nils in the string

function remove_nil(s)

    if s == nil then 
		return "" 
	else 
		return s 
	end
	
end


-- generating path to the specified file based on OS type

function generate_file_path(filename) 

	local userdata_dir = vlc.config.userdatadir()
	local conf_file = ""
	
	if string.match(userdata_dir, "^(%a:.+)$") then
		conf_file = userdata_dir .. "\\" .. filename
	else
		conf_file = userdata_dir .. "/" .. filename
	end
	
	return conf_file
	
end


-- helper function to see if the file exists

function file_exists(name)

	local f = io.open(name,"r")
	
	if f ~= nil then 
		io.close(f) 
		return true 
	else 
		return false 
	end
	
end


-- parsing config file to retrieve user credentials

function parse_config(filename)
	
	if not file_exists(filename) then
		return "", ""
	end

	local f = io.open(filename, "rb")
	local f_content = f:read "*a"
	f:close()
	
	local user_name = ""
	local user_pass = ""
	
	-- regexing username and password
	user_name, user_pass = f_content:match("([^,]+),([^,]+)")
	
	return remove_nil(user_name), remove_nil(user_pass)
	
end


-- reading session file

function parse_session(filename)

	if not file_exists(filename) then
		return ""
	end

	local f = io.open(filename, "rb")
	local f_content = f:read "*a"
	f:close()
	
	return remove_nil(f_content)
	
end


-- clearing current playlist

function clear_playlist()
	
	-- current item's id
	local current_id = math.floor(tonumber(vlc.playlist.current()))

	-- if video is playing, then saving this item
	if current_id > 0 then
		ITEM_TO_REMOVE = current_id
	end

	-- iterating through every item in the current playlist
	for _, v in pairs(vlc.playlist.get("playlist",false).children) do	
	
		local id = math.floor(tonumber(v.id))
		if id ~= current_id then -- deleting every item except for the one that is currently playing
			vlc.playlist.delete(v.id)
		end
		
	end

end


-- returning Flussonic token

function get_token()

	-- retrieving path to the session file
	local session_file = generate_file_path(FILE_SESSION)

	-- reading current session id
	session_id = parse_session(session_file)

	-- we are running low on tokens, so have to request some more from the API
	if #TOKENS < 100 then
	
		local string = ""
	
		-- trying to query the API's random token endpoint
		local response = vlc.stream(API_URL .. "?sessionId=" .. session_id .. "&apiAction=getRandomTokens&cnt=500")
		
		while true do
			local line = response:read(128)
			if line == nil then 
				break 
			end
		string = string .. line
		end

		local api_resp = json.decode(string)
		
		-- iterating through the list of tokens from API response
		for _, v in ipairs(api_resp.data.tokens) do
			table.insert(TOKENS, v) -- saving each of them as we go through the list
		end
	
	end
	
	-- removing a single token and returning it
	return table.remove(TOKENS)

end


-- checking for account validity

function check_account()

	local is_account_valid = false
	
	-- retrieving path to the config file
	local conf_file = generate_file_path(FILE_CONFIG)
	local session_file = generate_file_path(FILE_SESSION)
	
	-- if config can't be found, then we obviously have no account settings to check
	if not file_exists(conf_file) then
		return is_account_valid
	end
	
	-- otherwise let's try and retrieve users credentials from the config file
	
	local user_name = ""
	local user_pass = ""
	
	user_name, user_pass = parse_config(conf_file)
	
	local string = ""
	
	-- trying to query the API's auth endpoint
	local response = vlc.stream(API_URL .. "?userLogin=" .. user_name .. "&userPasswd=" .. md5.sumhexa(user_pass))

	while true do
		local line = response:read(128)
		if line == nil then 
			break 
		end
		string = string .. line
	end

	local api_resp = json.decode(string)
	
	-- are credentials correct?
	if api_resp.status ~= 1 then
		is_account_valid = false
	else
		is_account_valid = true
		file = io.open(session_file, "w")
		file:write(api_resp.data.sessionId)
		file:close()
		get_token()
		IS_SESSION_OPENED = true
	end
	
	return is_account_valid
	
end


-- helper for iterating through a table in ordered fashion

function __genOrderedIndex( t )
    local orderedIndex = {}
    for key in pairs(t) do
        table.insert( orderedIndex, key )
    end
    table.sort( orderedIndex )
    return orderedIndex
end

function orderedNext(t, state)
    -- Equivalent of the next function, but returns the keys in the alphabetic
    -- order. We use a temporary ordered key table that is stored in the
    -- table being iterated.

    local key = nil
    --print("orderedNext: state = "..tostring(state) )
    if state == nil then
        -- the first time, generate the index
        t.__orderedIndex = __genOrderedIndex( t )
        key = t.__orderedIndex[1]
    else
        -- fetch the next value
        for i = 1,#t.__orderedIndex do
            if t.__orderedIndex[i] == state then
                key = t.__orderedIndex[i+1]
            end
        end
    end

    if key then
        return key, t[key]
    end

    -- no more value to return, cleanup
    t.__orderedIndex = nil
    return
end

function orderedPairs(t)
    -- Equivalent of the pairs() function on tables. Allows to iterate
    -- in order
    return orderedNext, t, nil
end


-- display dialog for the configuration window

function create_dialog_config()	

	IS_CONFIG_DIALOG_OPEN = true
	
	-- retrieving path to the config file
	local conf_file = generate_file_path(FILE_CONFIG)
	
	local user_name = ""
	local user_pass = ""
	
	-- reading credentials
	user_name, user_pass = parse_config(conf_file)
	
	-- generating widget and its elements
	dlg = vlc.dialog("TvTeam Config")
   
	label_username = dlg:add_label("Ваш логин:", 1, 1, 2, 1)
	input_username = dlg:add_text_input(user_name, 2, 1, 4, 1)

	label_password = dlg:add_label("Ваш пароль:", 1, 2, 2, 1)
	input_password = dlg:add_text_input(user_pass, 2, 2, 4, 1)
	
	label_status = dlg:add_label("", 1, 3, 6, 1)
	label_line = dlg:add_label("_______________________________________________", 1, 4, 6, 1)
	
	button_save = dlg:add_button("Сохранить", action_Save, 1, 6, 1, 1)
	button_close = dlg:add_button("Закрить", action_Close, 2, 6, 1, 1)
	
end


-- action function for configuration dialog's Save button

function action_Save()

	-- retrieving path to the config file
	local conf_file = generate_file_path(FILE_CONFIG)

	-- getting values from the input fields
	local input_username_value = input_username:get_text()
	local input_password_value = input_password:get_text()
	
	-- saving these values to the config file
	file = io.open(conf_file, "w")
	file:write(input_username_value .. "," .. input_password_value)
	file:close()
	
	-- trying to obtain a new session id
	
	if check_account() then
		label_status:set_text("Добро пожаловать! Можете закрыть... ")
	else
		label_status:set_text("Ваши данные неправильные! Введите заново...")
		IS_SESSION_OPENED = false
	end


end


-- action function for configuration dialog's Close button

function action_Close()

	close()

end


-- display dialog for the main window

function create_dialog_main()	

	IS_CONFIG_DIALOG_OPEN = false
	
	-- retrieving path to the session file
	local session_file = generate_file_path(FILE_SESSION)
	
	-- reading current session id
	session_id = parse_session(session_file)

	-- trying to query the API's user channels endpoint and receive JSON response
	local response = vlc.stream(API_URL .. "?sessionId=" .. session_id .. "&apiAction=getUserChannels&resultType=tree")

	local string = ""
	
	while true do
		local line = response:read(128)
		if line == nil then 
			break 
		end
		string = string .. line
	end

	-- converting received JSON string into a table
	local api_resp = json.decode(string)
	
	
	-- initializing dialog

	dlg = vlc.dialog("TvTeam IPTV")


	----- live section of the window
   
	dlg:add_label("* ЕФИР *", 1, 1, 2, 1)
	label_live = dlg:add_label("Выберите категорию для прямого ефира", 2, 1, 2, 1)
	dlg:add_label(" ", 3, 1, 2, 1)

	label_live_category_label = dlg:add_label("Категория:", 1, 2, 2, 1)
	label_live_category_dropdown = dlg:add_dropdown(2, 2, 4, 1)
	
	-- iterating through categories
	for _, v in orderedPairs(api_resp.data.userChannelsTree) do
	
		-- skipping favourite channel group
		if math.floor(tonumber(v.groupId)) > 0 then 
			label_live_category_dropdown:add_value(v.groupName, v.groupId)
		end
	
	end
	
	button_live_all = dlg:add_button("Плейлист для всех каналов", function() action_Live_Channels("all") end, 1, 3, 1, 1)
	button_live_cat = dlg:add_button("Плейлист для категории", function() action_Live_Channels(label_live_category_dropdown:get_value()) end, 2, 3, 1, 1)
	
	dlg:add_label("-----------------------------------------------", 1, 4, 6, 1)
	
	
	----- archive section of the window
	
	dlg:add_label("* АРХИВ *", 1, 5, 2, 1)
	label_live = dlg:add_label("Выберите канал для просмостра архива", 2, 5, 2, 1)
	dlg:add_label(" ", 3, 5, 2, 1)
	
	label_archive_channel_label = dlg:add_label("Канал:", 1, 6, 2, 1)
	label_archive_channel_dropdown = dlg:add_dropdown(2, 6, 4, 1)
	
	-- iterating through categories
	for _, v in orderedPairs(api_resp.data.userChannelsTree) do
	
		-- skipping favourite channel group
		if math.floor(tonumber(v.groupId)) > 0 then 
		
			-- going through the channels in this group
			for __, t in orderedPairs(v.channelsList) do
			
				-- making sure this channel has EPG data available
				if string.len(remove_nil(t.curProgram.prTitle)) > 0 then
					label_archive_channel_dropdown:add_value(t.channelName, t.channelId)
				end
			
			end
			
		end
	
	end
	
	dlg:add_label("Дата в архиве:", 1, 7, 2, 1)
	label_archive_channel_date = dlg:add_dropdown(2, 7, 1, 1)
	button_archive = dlg:add_button("Плейлист для даты", 
		function() 
			action_Archive(label_archive_channel_dropdown:get_value(), label_archive_channel_date:get_value()) 
		end, 
	3, 7, 1, 1)
	
	-- max amount of days we can go back in the archive
	local archive_size = 5
	local days_for_subtraction = 0
	
	local date_today = date() -- current datetime in local time
	local date_epoch = date.epoch() -- epoch datetime
	
	local arch_date
	local arch_date_epoch
	local date_fmt
	local date_fmt_timestamp
	
	-- looping and formating dates
	for i=0,archive_size do
	
		days_for_subtraction = 0 - i -- calculating days to be substracted
		arch_date = date_today:copy():adddays(days_for_subtraction):sethours(0, 0, 0) -- creating archive date
		date_fmt = arch_date:fmt("%Y-%m-%d") -- formatting this archive date as human readable string
		
		arch_date_epoch = arch_date:toutc() - date_epoch -- getting time difference between archive date and epoch time
		date_fmt_timestamp = arch_date_epoch:spanseconds() -- converting it to a timestamp utc value
	
		label_archive_channel_date:add_value(date_fmt, date_fmt_timestamp)
	
	end
	
	
	----- buttons at the bottom of the window
   
	dlg:add_label("_______________________________________________", 1, 10, 6, 1)
	
	button_save = dlg:add_button("Конфиг", action_Config, 1, 11, 1, 1)
	button_close = dlg:add_button("Закрить", action_Close, 2, 11, 1, 1)
	
end


-- action function for main dialog's Live All and Live Category buttons

function action_Live_Channels(cat)

	clear_playlist()

	-- retrieving path to the session file
	local session_file = generate_file_path(FILE_SESSION)
	
	-- reading current session id
	session_id = parse_session(session_file)
	
	-- trying to query the API's user channels endpoint
	local response = vlc.stream(API_URL .. "?sessionId=" .. session_id .. "&apiAction=getUserChannels&resultType=tree")

	local string = ""
	
	while true do
		local line = response:read(128)
		if line == nil then 
			break 
		end
		string = string .. line
	end

	local api_resp = json.decode(string)
		
	local playlist_items = {}
	
	-- iterating through categories
	for _, v in orderedPairs(api_resp.data.userChannelsTree) do
	
		-- skipping favourite channels
		if math.floor(tonumber(v.groupId)) > 0 then
		
			-- to iterate through the channels in this group we either have to request all groups or specify one exact group
			if (tostring(cat) == "all" or tostring(cat) == tostring(v.groupId)) then
			
				-- iterating through all the channels in this group
				for __, t in orderedPairs(v.channelsList) do
				
					-- initializing playlist item for this channel
					local playlist_item = {}
					
					-- parsing the channel data
					
					if string.len(remove_nil(t.curProgram.prTitle)) > 0 then -- EPG data is available
						playlist_item.name = t.channelName .. " [" .. t.curProgram["%"] .. "%, " .. t.curProgram.prTitle .. "]"
					else -- no EPG is currently available for this channel, so just showing the channel name
						playlist_item.name = t.channelName
					end
					
					playlist_item.path = t.liveLink .. "?token=" .. get_token()
					
					table.insert(playlist_items, playlist_item) -- adding this element to the main playlist table
					
				end
				
			end
		
		end
	
	end
	
	-- adding this playlist to VLC
	vlc.playlist.enqueue(playlist_items)
	
	-- removing top element from the playlist when it's a residual from the last playlist
	if ITEM_TO_REMOVE > 0 then
		vlc.playlist.next()
		vlc.playlist.delete(ITEM_TO_REMOVE)
		vlc.playlist.stop()
	end

end


-- action function for main dialog's Archive button

function action_Archive(chid, day_begins)

	clear_playlist()

	-- retrieving path to the session file
	local session_file = generate_file_path(FILE_SESSION)
	
	-- reading current session id
	session_id = parse_session(session_file)
	
	-- trying to query the API's user channels endpoint
	local response = vlc.stream(API_URL .. "?sessionId=" .. session_id .. "&apiAction=getTvProgram&channelId=" .. tostring(chid))

	local string = ""
	
	while true do
		local line = response:read(128)
		if line == nil then 
			break 
		end
		string = string .. line
	end

	local api_resp = json.decode(string)
		
	local playlist_items = {}

	-- calculating max unixtime for event's beginning
	local day_ends = math.floor(tonumber(day_begins)) + 86400 -- 60 seconds * 60 minutes * 24 hours

	local event_start, event_finish
	local event_start_fmt, event_finish_fmt
	local dt_day, dt_month, dt_year, dt_hour, dt_min, dt_sec
	local event_duration, event_duration_sec
	local stream_url
	
	-- iterating through channel's EPG items
	for _, v in orderedPairs(api_resp.data.tvProgram) do
	
		-- accepting all events except the ones that are set to begin in the future
		if (v.timeMarker ~= "future") then
		
			-- making sure the event begins between 00:00:00 and 23:59:59
			if (math.floor(v.prStartSec) >= day_begins and math.floor(v.prStartSec) < day_ends) then
				
				-- matching and parsing datetimes from API response
				dt_day, dt_month, dt_year, dt_hour, dt_min, dt_sec = v.prStart:match("(%d+).(%d+).(%d+) (%d+):(%d+):(%d+)")
				event_start = date(dt_year .. "-" .. dt_month .. "-" .. dt_day .. " " .. dt_hour .. ":" .. dt_min .. " " .. TIMEZONE_IN_API_RESPONSE):tolocal()
				
				dt_day, dt_month, dt_year, dt_hour, dt_min, dt_sec = v.prStop:match("(%d+).(%d+).(%d+) (%d+):(%d+):(%d+)")
				event_finish = date(dt_year .. "-" .. dt_month .. "-" .. dt_day .. " " .. dt_hour .. ":" .. dt_min .. " " .. TIMEZONE_IN_API_RESPONSE):tolocal()
				
				-- formating these datetimes as readable strings
				event_start_fmt = event_start:fmt("%H:%M")
				event_finish_fmt = event_finish:fmt("%H:%M")
				
				-- calculating event's length in seconds
				event_duration = event_finish - event_start
				event_duration_sec = tostring(event_duration:spanseconds())
				
				-- generating URL for the stream by replacing NOW with event duration on a live event or simply adding token on the past ones
				stream_url = string.gsub(v.streamLink, "(-now.m3u8)", "-" .. event_duration_sec .. ".m3u8") .. "?token=" .. get_token()
				
				local playlist_item = {}
				
				playlist_item.name = "[" .. event_start_fmt .. "-" .. event_finish_fmt .. "] " .. v.prTitle
				playlist_item.path = stream_url
				
				-- adding this element to the main playlist table
				table.insert(playlist_items, playlist_item)
				
			end
			
		end
	
	end

	-- adding this playlist to VLC
	vlc.playlist.enqueue(playlist_items)
	
	-- removing top element from the playlist when it's a residual from the last playlist
	if ITEM_TO_REMOVE > 0 then
		vlc.playlist.next()
		vlc.playlist.delete(ITEM_TO_REMOVE)
		vlc.playlist.stop()
	end

end


-- action function for main dialog's Config button

function action_Config()

	close_dialog()
	create_dialog_config()

end


